from bq_iterate.core import BqQueryRowIterator, BqTableRowIterator
from bq_iterate.utils import batchify_iterator

__version__ = "0.1.0"

__all__ = ["BqQueryRowIterator", "BqTableRowIterator", "batchify_iterator"]



import logging
from google.cloud import bigquery

from bq_iterate.retry import default_retry

class BqRowIterator:
    client = None

    def __init__(self,  batch_size=10000, client=None):
        self.batch_size = batch_size
        self.initialized = False
        self.nth_batch = 0
        self.__class__.client = bigquery.Client() \
            if ((client is None) or (self.__class__.client is None)) else client

    def __iter__(self):
        return self

    def __next__(self):
        if not self.initialized:
            self.initialize()
            return next(self.row_iterator)
        if (el := next(self.row_iterator, None)) :
            return el
        elif self.row_iterable.next_page_token is None:
            raise StopIteration
        else:
            self.call_next_batch()
            return next(self.row_iterator)

    @default_retry
    def call_next_batch(self):
        page_token = (
            self.row_iterable.next_page_token if hasattr(self, "row_iterable") else None
        )
        logging.info(
            f"""calling bq client list_rows with batch size : {self.batch_size}
            batch number :{self.nth_batch} with page_token :{page_token or "None"}"""
        )
        self.row_iterable = self.list_next_rows(page_token)
        self.row_iterator = iter(self.row_iterable)
        self.nth_batch += 1

    @classmethod
    def update_client(cls, client):
      cls.__class__.client = client

    def list_next_rows(self, page_token= None):
        pass

    def initialize(self):
        self.call_next_batch()
        self.initialized = True


class BqQueryRowIterator(BqRowIterator):

    def __init__(self,  query, batch_size=10000, client=None):
        self.query = query
        super().__init__(batch_size,  client=client)
        self.query_job = self.client.query(query)

    def list_next_rows(self, page_token= None):
        return self.client.list_rows(
            self.query_job.destination,
            max_results=self.batch_size,
            page_token=page_token,
        )


class BqTableRowIterator(BqRowIterator):

    def __init__(self, project_id, dataset_id, table_id, batch_size=10000, client=None):
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        super().__init__(batch_size=batch_size, client=client)

    def list_next_rows(self, page_token=None):
        try:
            return self.client.list_rows(
                f"{self.project_id}.{self.dataset_id}.{self.table_id}",
                max_results=self.batch_size,
                page_token=page_token,
            )
        except Exception as e:
          logging.error(f'something wrong happend : {e}')
          raise



from bq_iterate import config

from tenacity import (
    retry as _retry,
    wait_exponential,
    before_sleep_log,
    before_log,
    stop_after_attempt,
)


def default_retry(fn):
    return _retry(
        wait=wait_exponential(
            multiplier=config.RETRY_CONF.get("wait_multiplier", 2),
            min=config.RETRY_CONF.get("wait_min", 4),
            max=config.RETRY_CONF.get("wait_max", 20),
        ),
        stop=stop_after_attempt(config.RETRY_CONF.get("stop_after_attempt", 5)),
        reraise=True,
        before=before_log(config.logger, config.logging.INFO),
        before_sleep=before_sleep_log(config.logger, config.logging.INFO),
    )(fn)


tn_retry = _retry



import sys
import logging


FORMAT = "%(asctime)s - (%(name)s) - %(levelname)s - %(message)s"

logging.basicConfig(format=FORMAT)

logger = logging.getLogger()
logger.setLevel(logging.ERROR)

handler = logging.StreamHandler(sys.stdout)
for handler in logger.handlers:
    logger.removeHandler(handler)

logger.addHandler(handler)

RETRY_CONF = {
    "wait_multiplier": 2,
    "wait_min": 4,
    "wait_max": 40,
    "stop_after_attempt": 6,
}



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bq_iterate']

package_data = \
{'': ['*']}

install_requires = \
['google-cloud-bigquery>=2.2.0,<3.0.0', 'tenacity>=7.0.0,<8.0.0']

setup_kwargs = {
    'name': 'bq-iterate',
    'version': '0.1.4',
    'description': 'Work with BigQuery data as you do with lists in python',
    'long_description': '# Introduction\nThis project serves as a BigQuery helper that allows you to query data from BigQuery, without worrying about memory limitation concerns, it makes working with BigQuery data as easy as working with lists in python\n\n## Installation : \n```pip install bq-iterate```\n## Usage\n```python\nfrom bq_iterate import BqQueryRowIterator, batchify_iterator\nquery = "select * from <project_id>.<dataset_id>.<table_id>"\nrow_itrator = BqQueryRowIterator(query=query, batch_size=2000000) #  choose a batch_size that will fit into your memory\nbatches = batchify_iterator(row_itrator, batch_slice=50000) #  choose a batch_slice that will fit into your memory\ndata = []\nfor batch in batches:\n    # do your batch processing here\n    data.append(len(batch))\nprint(sum(data))\n```\n\n## What happens behind the scenes :\n\n**bq_iterate provide two functionalities\xa0:**\n\n* 2 classes BqQueryIterator and BqTableRowIterator, they behave like an iterator, where they hold only <batch_size> elements in memory and when you want to access the element <batch_size + 1> the iterator calls in memory the next batch_size + 1 elements\n\n* A function batchify_iterator, what this function does, it takes an iterator and yields slices of it, the <batch_slice> can be bigger than the <batch_size> even if by common sens it\'s supposed to be smaller, it doesn\'t matter, since batchify_iterator will create in memory at each batch it yields, a list of <batch_slice> elements, once the batch consumed it freed from memory, since it\'s a generator\n',
    'author': 'Senhaji-Rhazi-Hamza',
    'author_email': 'hamza.senhajirhazi@gmal.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Senhaji-Rhazi-Hamza/bq_iterate',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)




from collections.abc import Iterable, Iterator

from bq_iterate.config import logging


def batchify_iterator(iterator, batch_slice=100):

    batchify_iterator.batch_count = 0

    if isinstance(iterator, Iterable) and not isinstance(iterator, Iterator):
        iterator = iter(iterator)
    elif not isinstance(iterator, Iterator):
        raise Exception(f"{type(iterator)} not iterable")

    while (el := next(iterator, None)) is not None:

        batchify_iterator.batch_count += 1

        logging.info(
            f"""yielding the {batchify_iterator.batch_count}th batch
            for a batch slice {batch_slice}"""
        )
        yield [el] + \
              [
                  el for _ in range(batch_slice - 1)
                  if (el := next(iterator, None))
              ]
    batchify_iterator.batch_count = 0

def bq_row_to_dict(bq_row):
    return {k: v for k, v in bq_row.items()}
